var ModelReport;

$(function() {
	ModelReport = (function() {
		var planSvg = d3.select('#planSvg');
		var elevationSvg = d3.select('#elevationSvg');
		var normalLineGroup = planSvg.append('g');
		var intersectionLineGroup = planSvg.append('g');
		var minModelHeight, maxModelHeight;
		var minScreenHeight, maxScreenHeight;

		var flippedXY = false;
		function getPolygonBounds(polygons) {
			var bounds = [{x: Infinity, y: Infinity, z: Infinity}, {x: -Infinity, y: -Infinity, z: -Infinity}];

			polygons.forEach(function(polygon) {
				polygon.points.forEach(function(point) {
					bounds[0].x = Math.min(bounds[0].x, point.x);
					bounds[0].y = Math.min(bounds[0].y, point.y);
					bounds[0].z = Math.min(bounds[0].z, point.z);

					bounds[1].x = Math.max(bounds[1].x, point.x);
					bounds[1].y = Math.max(bounds[1].y, point.y);
					bounds[1].z = Math.max(bounds[1].z, point.z);
				});
			});

			return bounds;
		}

		function polygonWithPoints(polygon, newPoints) {
			return {
				id: polygon.id,
				color: polygon.color,
				tooltipHtml: polygon.tooltipHtml,
				errors: polygon.errors,
				points: newPoints
			}
		}

		function mapPolygonPoints(polygons, f) {
			return polygons.map(function(polygon) {
				return polygonWithPoints(polygon, polygon.points.map(f));
			});
		}

		function flatMapPolygonPointsWithPrev(polygons, f) {
			var newPolygons = [];
			polygons.forEach(function(polygon) {
				var points = polygon.points;
				var newPoints = [];
				for(var i = 0; i < points.length; i++) {
					var point1 = points[i];
					var point2 = points[(i + 1) % points.length];
					f.call(null, point1, point2).forEach(function(x) { newPoints.push(x) });
				}
				if(newPoints.length > 0) {
					newPolygons.push(polygonWithPoints(polygon, newPoints));
				}
			});
			return newPolygons;
		}

		function fitPolygons(polygons) {
			var bounds = getPolygonBounds(polygons);

			if(!flippedXY) {
				if(bounds[1].x - bounds[0].x < bounds[1].y - bounds[0].y) {
					polygons = mapPolygonPoints(polygons, function(point) { return {x: -point.y, y: point.x, z: point.z} });
					bounds = getPolygonBounds(polygons);
				}
			}

			var sx = 800 / (bounds[1].x - bounds[0].x);
			var sy = 400 / (bounds[1].y - bounds[0].y);
			var sz = 400 / (bounds[1].z - bounds[0].z);

			var s = Math.min(sx, sy, sz);

			var ox = bounds[1].x - (bounds[1].x - bounds[0].x) / 2;
			var oy = bounds[1].y - (bounds[1].y - bounds[0].y) / 2;
			var oz = bounds[1].z - (bounds[1].z - bounds[0].z) / 2;

			return mapPolygonPoints(polygons, function(point) {
				return {
					x: (point.x - ox) * s + 800 / 2,
					y: (point.y - oy) * s + 400 / 2,
					z: (point.z - oz) * s + 400 / 2
				};
			});
		}

		function intersect(point1, point2, z) {
			var x1 = point1.x, y1 = point1.y, z1 = point1.z;
			var x2 = point2.x, y2 = point2.y, z2 = point2.z;

			if(z2 == z1) return null;

			var t = (z - z1) / (z2 - z1);

			if(t < 0 || t > 1) return null;

			var x = x1 + t * (x2 - x1);
			var y = y1 + t * (y2 - y1);

			return {x: x, y: y, z: z};
		}

		function clipPolygons(polygons, height) {
			return flatMapPolygonPointsWithPrev(polygons, function(point1, point2) {
				var newPoints = [];

				if(point1.z <= height) {
					newPoints.push(point1);
				}

				if((point1.z < height) != (point2.z < height)) {
					var intersectionPoint = intersect(point1, point2, height);
					if(intersectionPoint) {
						newPoints.push(intersectionPoint);
					}
				}

				return newPoints;
			});
		}

		function getIntersectingLines(polygons, height) {
			var lines = [];

			polygons.forEach(function(polygon) {
				for(var i = 0; i < polygon.points.length; i++) {
					var point1 = polygon.points[i];
					var point2 = polygon.points[(i + 1) % polygon.points.length];

					if(point1.z == height && point2.z == height) {
						lines.push([point1, point2]);
					}
				}
			});

			return lines;
		}

		function addPlan(polygons, height) {
			function tx(d) { return d }
			function ty(d) { return 400 - d }

			var line = d3.svg.line()
				.x(function(d) { return tx(d.x) })
				.y(function(d) { return ty(d.y) });

			var clippedPolygons = clipPolygons(polygons, height);

			// Add normal lines.
			paths = normalLineGroup.selectAll('path')
				.data(clippedPolygons);

			paths.enter().append('path')
				.on('mouseover', function(d) {
					if(d.tooltipHtml) {
						var tooltip = d3.select(".tooltip");
						tooltip.transition().duration(200).style('opacity', 0.95);
						tooltip.html(d.tooltipHtml)
							.style('left', (d3.event.pageX) + 'px')
							.style('top', (d3.event.pageY - 28) + 'px');
					}
				})
				.on('mouseout', function(d) {
					var tooltip = d3.select(".tooltip");
					tooltip.transition().duration(500).style('opacity', 0);
				});

			paths.exit().remove();

			paths.attr('stroke', '#aaa')
				.attr('fill', function(d) { return d.color ? d.color : 'white' })
				.attr('d', function(d) { return line(d.points) + "Z" })
				
			// Add bold intersection lines.
			lines = intersectionLineGroup.selectAll('line')
				.data(getIntersectingLines(clippedPolygons, height));

			lines.enter().append('line');
			lines.exit().remove();

			lines.attr('stroke', 'black')
				.attr('stroke-width', 2)
				.attr('x1', function(d) { return tx(d[0].x) })
				.attr('y1', function(d) { return ty(d[0].y) })
				.attr('x2', function(d) { return tx(d[1].x) })
				.attr('y2', function(d) { return ty(d[1].y) });
		}

		function addElevation(polygons, height) {
			function tx(d) { return d }
			function ty(d) { return 400 - d }

			var line = d3.svg.line()
				.x(function(d) { return flippedXY ? tx(d.z) : tx(d.y) })
				.y(function(d) { return flippedXY ? ty(d.y) : ty(d.z) });

			var clippedPolygons = clipPolygons(polygons, height);
			paths = elevationSvg.selectAll('path')
				.data(clippedPolygons);

			paths.enter()
				.append('path')
				.attr('stroke', 'black')
				.attr('fill', 'transparent')

			paths.exit().remove();

			paths.attr('d', function(d) { return line(d.points) + "Z" });

			elevation = elevationSvg.selectAll('line')
				.data([height]);

			elevation.enter().append('line');
			elevation.exit().remove();

			elevation.attr('stroke', 'red')
				.attr(flippedXY ? 'y1' : 'x1', 0)
				.attr(flippedXY ? 'y2' : 'x2', 400)
				.attr(flippedXY ? 'x1' : 'y1', function(d) { return flippedXY ? tx(d) : 400 - tx(d) })
				.attr(flippedXY ? 'x2' : 'y2', function(d) { return flippedXY ? tx(d) : 400 - tx(d) });
		}

		function getModelHeight(z) {
			return (z - minScreenHeight) / (maxScreenHeight - minScreenHeight) * (maxModelHeight - minModelHeight) + minModelHeight;
		}

		function onSlide(height, polygons) {
			addPlan(polygons, height);
			addElevation(polygons, height);
			$("#heightValue").text(Math.floor(getModelHeight(height) * 10) / 10);
		}

		function initSlider() {
			$("#slider").slider();
			$("#yzSelect").prop('checked', true);
		}

		function refreshSlider(polygons) {
			$("#slider").slider("option", {
				min: minScreenHeight + 1,
				max: maxScreenHeight,
				value: minScreenHeight + 1,
				step: 0.1
			});

			$("#slider").off("slide");
			$("#slider").on("slide", function(event, ui) { onSlide(ui.value, polygons); });

			onSlide(minScreenHeight + 1, polygons);
		}

		function initViewSelect(polygons) {
			$("#viewSelect").buttonset();

			$("#xySelect").click(function() {
				var newPolygons = mapPolygonPoints(polygons, function(point) { return { x: point.x, y: point.y, z: point.z } });
				flippedXY = false;
				setPolygons(newPolygons);
			});

			$("#yzSelect").click(function() {
				var newPolygons = mapPolygonPoints(polygons, function(point) { return { x: point.y, y: point.z, z: point.x } });
				flippedXY = true;
				setPolygons(newPolygons);
			});

			$("#xzSelect").click(function() {
				var newPolygons = mapPolygonPoints(polygons, function(point) { return { x: point.x, y: point.z, z: point.y } });
				flippedXY = true;
				setPolygons(newPolygons);
			});
		}

		function setPolygons(polygons) {
			polygons = _.sortBy(polygons, function(polygon) {
				var heights = polygon.points.map(function(point) { return point.z });
				return _.max(heights);
			});

			var errors = polygons.reduce(function(errors, polygon) { return polygon.errors + errors }, 0);
			$("#errors").text("Errors: " + errors);

			var bounds = getPolygonBounds(polygons);
			minModelHeight = bounds[0].z;
			maxModelHeight = bounds[1].z;

			polygons = fitPolygons(polygons);

			var bounds = getPolygonBounds(polygons);
			minScreenHeight = Math.floor(bounds[0].z + 1);
			maxScreenHeight = Math.ceil(bounds[1].z);

			refreshSlider(polygons);

			d3.select("body").append("div")
				.attr("class", "tooltip")
				.style("opacity", 0);
		}

		function setInitialPolygons(polygons) {
			initViewSelect(polygons);
			initSlider();
			setPolygons(polygons, false);
		}

		return {
			setInitialPolygons: setInitialPolygons
		};
	})();

	window.location = 'skp:loadComplete';
});

function setPolygons(polygons) {
	ModelReport.setInitialPolygons(polygons);
}